#include <fstream>
#include <cmath>
#include <cassert>
#include "FeatureSet.h"

static int iround ( float x )
{
  if ( x < 0.0 )
  {
    return ( int )( x - 0.5 );
  }
  else
  {
    return ( int )( x + 0.5 );
  }
}

// Create a feature.
Feature::Feature ()
{
  selected = false;
}

// Write the feature to an output stream.
ostream & operator<< ( ostream & os, const Feature & f )
{
  os << f.type << '\n';
  assert ( f.id );
  os << f.id << '\n';
  os << f.x << ' ' << f.y << '\n';
  os << f.angleRadians << '\n';

  os << f.data.size () << '\n';

  for ( unsigned int i = 0; i < f.data.size (); i++ )
  {
    os << f.data[i] << '\n';
  }

  return os;
}

// Read the feature from an input stream.
istream & operator>> ( istream & is, Feature & f )
{
  int n;

  is >> f.type;
  is >> f.id;
  assert ( f.id );
  is >> f.x >> f.y;
  is >> f.angleRadians;

  is >> n;

  f.data.clear ();
  f.data.resize ( n );

  for ( int i = 0; i < n; i++ )
  {
    is >> f.data[i];
  }

  return is;
}

// Create a feature set.
FeatureSet::FeatureSet()
{
}

// Load a feature set from a file.
bool FeatureSet::load ( const char * name )
{
  int n;

  // Clear the currently loaded features.
  clear ();

  // Open the file.
  ifstream f ( name );

  if ( !f.is_open () )
  {
    return false;
  }

  // Read the total number of features.
  f >> n;

  // Resize the vector of features.
  resize ( n );

  // Read each of the features.
  iterator i = begin ();

  while ( i != end () )
  {
    f >> ( *i );
    i++;
  }

  // Close the file.
  f.close ();

  return true;
}

// Save a feature set to file.
bool FeatureSet::save ( const char * name ) const
{
  // Open the file.
  ofstream f ( name );

  if ( !f.is_open () )
  {
    return false;
  }

  // Write the number of features.
  f << size () << '\n';

  // Write each of the features.
  const_iterator i = begin ();

  while ( i != end () )
  {
    f << ( *i );
    i++;
  }

  // Close the file.
  f.close ();

  return true;
}

// Select (or deselect) features at a location.
void FeatureSet::select_point ( int x, int y )
{
  iterator i = begin ();

  while ( i != end () )
  {
    // If the given point is within 5 pixels of the feature, then
    // select it.  This can select multiple features if they are
    // nearly collocated.
    if ( ( abs ( ( *i ).x - x ) <= 5 ) && ( abs ( ( *i ).y - y ) <= 5 ) )
    {
      ( *i ).selected = ( !( *i ).selected );
      assert ( i->id );
      std::cout << "selecting feature " << i->id << " [orient: " <<
      i->angleRadians << "]" << std::endl;
    }

    i++;
  }
}

// Select (or deselect) features inside a box.
void FeatureSet::select_box ( int xMin, int xMax, int yMin, int yMax )
{
  iterator i = begin ();

  while ( i != end () )
  {
    if ( ( ( *i ).x >= xMin ) && ( ( *i ).x <= xMax ) &&
         ( ( *i ).y >= yMin ) && ( ( *i ).y <= yMax ) )
    {
      ( *i ).selected = ( !( *i ).selected );
    }

    i++;
  }
}

// Select all features.
void FeatureSet::select_all ()
{
  iterator i = begin ();

  while ( i != end () )
  {
    ( *i ).selected = true;
    i++;
  }
}

// Deselect all features.
void FeatureSet::deselect_all ()
{
  iterator i = begin ();

  while ( i != end () )
  {
    ( *i ).selected = false;
    i++;
  }
}

// Take only the selected features.
void FeatureSet::get_selected_features ( FeatureSet & f )
{
  f.clear ();

  iterator i = begin ();

  while ( i != end () )
  {
    if ( ( *i ).selected )
    {
      f.push_back ( ( *i ) );
    }

    i++;
  }
}
