#include "OpticalFlow.h"
#include <vector>
#include <math.h>

#define MAX_CORNERS 100
#define WINSIZE 10
#define LEVELS 5
#define MAX_FEATURE_ERROR 550
#define MIN_MATCHES 30
#define USE_FAST 1
#define FAST_TH 25
#define DISP_ERROR_VAR 0.1	//pixel
#define IMG_W 640
#define IMG_H 480

struct _matches{
	CvPoint2D32f f1;
	CvPoint2D32f f2;
	float score;
};

enum HEADING{
	NORTH,
	NORTHEAST,
	EAST,
	SOUTHEAST,
	SOUTH,
	SOUTHWEST,
	WEST,
	NORTHWEST
};

float deg2rad(float deg);
float rad2deg(float rad);
int get_distribution(std::vector<float>&input, float *mean, float *var);

OpticalFlow::OpticalFlow(CvSize size)
{
	curr_image = cvCreateImage(size,8,1);
	prev_image = cvCreateImage(size,8,1);

	eig_image = cvCreateImage(size,IPL_DEPTH_32F,1);
	tmp_image = cvCreateImage(size,IPL_DEPTH_32F,1);
	
	CvSize pyrSz = cvSize((size.width+8),(int)(size.height/3));
	pyr_prev = cvCreateImage(pyrSz,IPL_DEPTH_32F,1);
	pyr_curr = cvCreateImage(pyrSz,IPL_DEPTH_32F,1);

	corners_prev = new CvPoint2D32f [MAX_CORNERS];
	corners_curr = new CvPoint2D32f [MAX_CORNERS];
	corners_output = new CvPoint2D32f [MAX_CORNERS];

	memset(corners_prev,0,sizeof(CvPoint2D32f)*MAX_CORNERS);
	memset(corners_curr,0,sizeof(CvPoint2D32f)*MAX_CORNERS);
	memset(corners_output,0,sizeof(CvPoint2D32f)*MAX_CORNERS);

	winsize = cvSize(WINSIZE,WINSIZE);
	srand(time(NULL));

}

OpticalFlow::~OpticalFlow(void)
{
	cvReleaseImage(&curr_image);
	cvReleaseImage(&prev_image);
	cvReleaseImage(&eig_image);
	cvReleaseImage(&tmp_image);
	
	cvReleaseImage(&pyr_prev);
	cvReleaseImage(&pyr_curr);

	delete[]corners_prev;
	delete[]corners_curr;
	delete[]corners_output;
}

void OpticalFlow::ComputeFASTFeatures ( IplImage * image, FeatureSet & features )
{
	std::vector<cv::KeyPoint> keypoints;
	cv::FAST(image, keypoints, FAST_TH);

	features.clear ();
	features.reserve ( keypoints.size () );
	for ( size_t i = 0; i < keypoints.size (); ++i )
	{
		Feature f;
		f.x = keypoints[i].pt.x;
		f.y = keypoints[i].pt.y;
		f.type = 3;
		f.id = 0;
		f.angleRadians = keypoints[i].angle;
		features.push_back ( f );
	}
	
}
int OpticalFlow::InitFlow(IplImage *image)
{
	starttime = clock();
	currtime = starttime;
	corner_count_curr = MAX_CORNERS;
	cvCopy(image, curr_image);

	if(USE_FAST){
		corner_count_curr = GetFASTFeatures(image,corners_curr);
	}else{
		cvGoodFeaturesToTrack(
			image,
			eig_image,
			tmp_image,
			corners_curr,
			&corner_count_curr,
			0.01,
			5.0,
			0,
			3, 
			0,
			0.04
			);
	}

	//cvCopyImage(image,key_image);
	//cvCopyImage(image,curr_image);
	//memcpy(corners_curr,corners_key,sizeof(CvPoint2D32f)*MAX_CORNERS);
	//cvCopyImage(pyr_curr,pyr_key);
	printf("%d\n",corner_count_curr);
	return corner_count_curr;

}

int OpticalFlow::GetFlow2(IplImage *image, CvPoint2D32f *displacement_vector, CvPoint2D32f *velocity_vector, float *orientation)
{
	char features_found[MAX_CORNERS];
	float feature_errors[MAX_CORNERS];
	
	matches_t m;
	
	int status=0;	

	//update
	corner_count_prev = corner_count_curr;
	corner_count_curr = MAX_CORNERS;

	cvCopy(curr_image,prev_image);
	cvCopy(image, curr_image);
	//cvCopyImage(pyr_prev,pyr_curr);
	
	memcpy(corners_prev,corners_curr,sizeof(CvPoint2D32f)*MAX_CORNERS);

	prevtime = currtime;
	currtime = clock();
	deltatime = (currtime - prevtime)*0.001;

	if(USE_FAST){
		corner_count_curr = GetFASTFeatures(image,corners_curr);
	}
	else{
		cvGoodFeaturesToTrack(
			curr_image,
			eig_image,
			tmp_image,
			corners_curr,
			&corner_count_curr,
			0.01,
			5.0,
			0,
			3, 
			0,
			0.04
			);
	}

	cvCalcOpticalFlowPyrLK(
		prev_image,
		curr_image,
		pyr_prev,
		pyr_curr,
		corners_prev,
		corners_output,
		corner_count_curr,
		winsize,
		LEVELS,
		features_found,
		feature_errors,
		cvTermCriteria( CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, -3),
		0 );
	
	
	matches.clear();

	//get directional vector into bins
	for(int i=0;i<corner_count_curr;i++){
		if( features_found[i]==0 || feature_errors[i]>MAX_FEATURE_ERROR){
			continue;
		}
		m.f1 = corners_prev[i];
		m.f2 = corners_output[i];
		m.score = feature_errors[i];
		matches.push_back(m);
	}
	
	if(matches.size()>10){
		status = RansacFlow(matches, displacement_vector, orientation);
	}
	
	//filter displacement vector
	float disp_err = displacement_vector->x*displacement_vector->x + displacement_vector->y*displacement_vector->y;
	if(disp_err<DISP_ERROR_VAR){
		displacement_vector->x = 0;
		displacement_vector->y = 0;
	}	


	velocity_vector->x = displacement_vector->x/(deltatime+0.0000000001);
	velocity_vector->y = displacement_vector->y/(deltatime+0.0000000001);
	
	return matches.size();
}

int OpticalFlow::RansacFlow(std::vector<matches_t> &m, CvPoint2D32f *displacement_vector, float *orientation)
{
	
	const float max_error = 3;
	const int max_ransac = 100;
	std::vector<CvPoint2D32f> angledist;
	std::vector<CvPoint2D32f> angledist_norm;
	float angle;
	float dist;
	float dx,dy;
	int rand_index;
	CvPoint2D32f testpt;
	CvPoint2D32f sum = cvPoint2D32f(0,0);
	CvPoint2D32f sumsq = cvPoint2D32f(0,0);
	CvPoint2D32f mean = cvPoint2D32f(0,0);
	CvPoint2D32f var = cvPoint2D32f(0,0);
	int inliner_count=0;
	CvPoint2D32f dispvec;	//displacement vector 

	//generate a list of angle and distance for faster computational ransac
	for(int i=0;i<m.size();i++){
		dx = (matches[i].f1.x - matches[i].f2.x);
		dy = (matches[i].f1.y - matches[i].f2.y);
		angle = atan2(dy, dx);
		dist = sqrt(dx*dx+dy*dy);
		angledist.push_back( cvPoint2D32f(angle,dist));

		sum.x += angle;
		sum.y += dist;

		sumsq.x += angle*angle;
		sumsq.y += dist*dist;
	}

	//normalise
	mean.x = sum.x/(float)angledist.size();
	mean.y = sum.y/(float)angledist.size();
	var.x = sumsq.x/(float)angledist.size() - mean.x*mean.x;
	var.y = sumsq.y/(float)angledist.size() - mean.y*mean.y;
	
	for(int i=0;i<angledist.size();i++){
		CvPoint2D32f norm;
		//norm.x = (angledist[i].x-mean.x)/var.x;
		norm.x = angledist[i].x;
		norm.y = (angledist[i].y-mean.y)/var.y;
		angledist_norm.push_back(norm);
	}

	float error=0;
	float max_inliners=0;
	int max_index=0;

	for(int ransac_try=0;ransac_try<max_ransac;ransac_try++){
		rand_index = rand()%angledist_norm.size();
		testpt = angledist_norm[rand_index];
		inliner_count=0;
		for(int i=0;i<angledist.size();i++){
			dx = testpt.x-angledist_norm[i].x;
			dy = testpt.y-angledist_norm[i].y;

			if(abs(dx)>CV_PI){
				dx = dx - 2*CV_PI;
			}

			error = dx*dx + dy*dy;

			if(error<max_error){
				inliner_count++;
			}
		}
		if(inliner_count>max_inliners){
			max_inliners=inliner_count;
			max_index=rand_index;
		}
	}
	dispvec.x = matches[max_index].f1.x - matches[max_index].f2.x;
	dispvec.y = matches[max_index].f1.y - matches[max_index].f2.y;
	
	*displacement_vector = dispvec;
	*orientation = angledist[max_index].x;
	//*displacement = angledist[max_index].y;	
	
	return 0;
}


void OpticalFlow::ResetFlow()
{
	cvCopy(curr_image,prev_image);
}
void OpticalFlow::DrawFlow(IplImage *out_image)
{
	if(matches.size() == 0){
		return;
	}
	
	for(int i=0;i<matches.size();i++){
		/*cvLine(
			out_image,
			cvPoint( cvRound(matches[i].f1.x), cvRound(matches[i].f1.y) ),
			cvPoint( cvRound(matches[i].f2.x), cvRound(matches[i].f2.y) ),
			CV_RGB(255,0,0),
			1);*/
		cvLine(out_image,cvPoint((int)matches[i].f1.x,(int)matches[i].f1.y),cvPoint((int)matches[i].f2.x,(int)matches[i].f2.y),CV_RGB(255,0,0), 1);
	}
}

float deg2rad(float deg)
{
	float rad = deg/180.0*CV_PI;
	return rad;
}
float rad2deg(float rad)
{
	float deg = rad/CV_PI*180.0;
	return deg;
}


int get_distribution(std::vector<float>&input, float *mean, float *var)
{
	float sum=0;
	float sum_sq=0;
	float m;
	float v;
		
	if(input.size() == 0){
		return -1;
	}

	for(int i=0;i<input.size();i++){
		sum+=input[i];
		sum_sq+=input[i]*input[i];
	}
	m = sum/(float)input.size();
	v = (sum_sq/(float)input.size()) - m*m;
	*mean = m;
	*var = v;
	printf("n: %d\tmean: %3.3f\tvar: %3.3f\n",input.size(),*mean,*var);
	return 0;
}

int OpticalFlow::GetFASTFeatures(IplImage *image, CvPoint2D32f *pt)
{
	int i;
	FeatureSet ffast;
	ComputeFASTFeatures(image, ffast);
	for(i=0;i<ffast.size() && i< MAX_CORNERS;i++){
		pt[i] = cvPoint2D32f(ffast[i].x,ffast[i].y);
	}
	return i;
}


